@@ -1,9 +1,10 @@
 /* hdparm.c - Command line interface to get/set hard disk parameters */
 /*          - by Mark Lord (C) 1994-2008 -- freely distributable */
 #include <unistd.h>
-#include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
+#define __USE_GNU	/* for O_DIRECT */
+#include <string.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <ctype.h>
@@ -25,7 +26,7 @@
 
 extern const char *minor_str[];
 
-#define VERSION "v9.6"
+#define VERSION "v9.13"
 
 #ifndef O_DIRECT
 #define O_DIRECT	040000	/* direct disk access, not easily obtained from headers */
@@ -49,6 +50,7 @@ extern const char *minor_str[];
 
 char *progname;
 int verbose = 0;
+int prefer_ata12 = 0;
 static int do_defaults = 0, do_flush = 0, do_ctimings, do_timings = 0;
 static int do_identity = 0, get_geom = 0, noisy = 1, quiet = 0;
 static int do_flush_wcache = 0;
@@ -367,14 +369,26 @@ static __u64 get_lba_capacity (__u16 *idw)
 	return nsects;
 }
 
+static char *strip (char *s)
+{
+	char *e;
+
+	while (*s == ' ') ++s;
+	for (e = s; *e && *++e != ' ';);
+	*e = '\0';
+	return s;
+}
+
 static void dump_identity (__u16 *idw)
 {
 	int i;
 	char pmodes[64] = {0,}, dmodes[128]={0,}, umodes[128]={0,};
+	char *model = strip(strndup((char *)&idw[27], 40));
+	char *fwrev = strip(strndup((char *)&idw[23],  8));
+	char *serno = strip(strndup((char *)&idw[10], 20));
 	__u8 tPIO;
 
-	printf("\n Model=%.40s, FwRev=%.8s, SerialNo=%.20s",
-		(char *)&idw[27], (char *)&idw[23], (char *)&idw[10]);
+	printf("\n Model=%.40s, FwRev=%.8s, SerialNo=%.20s", model, fwrev, serno);
 	printf("\n Config={");
 	for (i = 0; i <= 15; i++) {
 		if (idw[0] & (1<<i))
@@ -936,7 +950,7 @@ static __u64 do_get_native_max_sectors (int fd, __u16 *id)
 			err = errno;
 			perror (" READ_NATIVE_MAX_ADDRESS failed");
 		} else {
-			max = ((r.lob.lbah << 16) | (r.lob.lbam << 8) | r.lob.lbal) + 1;
+			max = (((r.lob.dev & 0x0f) << 24) | (r.lob.lbah << 16) | (r.lob.lbam << 8) | r.lob.lbal) + 1;
 		}
 	}
 	errno = err;
@@ -1670,14 +1684,19 @@ open_ok:
 	}
 	if (get_powermode) {
 		__u8 args[4] = {ATA_OP_CHECKPOWERMODE1,0,0,0};
-		const char *state;
+		const char *state = "unknown";
 		if (do_drive_cmd(fd, args)
 		 && (args[0] = ATA_OP_CHECKPOWERMODE2) /* (single =) try again with 0x98 */
 		 && do_drive_cmd(fd, args)) {
 			err = errno;
-			state = "unknown";
 		} else {
-			state = (args[2] == 255) ? "active/idle" : "standby";
+			switch (args[2]) {
+				case 0x00: state = "standby";		break;
+				case 0x40: state = "NVcache_spindown";	break;
+				case 0x41: state = "NVcache_spinup";	break;
+				case 0x80: state = "idle";		break;
+				case 0xff: state = "active/idle";	break;
+			}
 		}
 		printf(" drive state is:  %s\n", state);
 	}
@@ -1733,6 +1752,17 @@ open_ok:
 			}
 		}
 	}
+	if (get_apmmode) {
+		id = get_identify_data(fd, id);
+		printf(" APM_level	= ");
+		if((id[83] & 0xc008) == 0x4008) {
+			if (id[86] & 0x0008)
+				printf("%u\n", id[91] & 0xff);
+			else
+				printf("off\n");
+		} else
+			printf("not supported\n");
+	}
 	if (get_acoustic) {
 		id = get_identify_data(fd, id);
 		if (id) {
@@ -1875,6 +1905,7 @@ static void usage_help (int rc)
 	" --Istdin          read identify data from stdin as ASCII hex\n"
 	" --Istdout         write identify data to stdout as ASCII hex\n"
 	" --make-bad-sector deliberately corrupt a sector directly on the media (VERY DANGEROUS)\n"
+	" --prefer-ata12    use 12-byte (instead of 16-byte) SAT commands when possible\n"
 	" --read-sector     read and dump (in hex) a sector directly from the media\n"
 	" --security-help   display help for ATA security commands\n"
 	" --verbose         display extra diagnostics from some commands\n"
@@ -2193,6 +2224,9 @@ get_longarg (void)
 	if (0 == strcasecmp(name, "verbose")) {
 		verbose = 1;
 		--num_flags_processed;	/* doesn't count as an action flag */
+	} else if (0 == strcasecmp(name, "prefer-ata12")) {
+		prefer_ata12 = 1;
+		--num_flags_processed;	/* doesn't count as an action flag */
 	} else if (0 == strcasecmp(name, "yes-i-know-what-i-am-doing")) {
 		i_know_what_i_am_doing = 1;
 		--num_flags_processed;	/* doesn't count as an action flag */
@@ -2302,7 +2336,7 @@ int main (int _argc, char **_argv)
 				case GET_SET_PARM('a',"filesystem-read-ahead",fsreadahead,0,2048);
 				case GET_SET_PARM('A',"look-ahead",lookahead,0,1);
 				case GET_SET_PARM('b',"bus-state",busstate,0,2);
-				case     SET_PARM('B',"power-management-mode",apmmode,1,255);
+				case GET_SET_PARM('B',"power-management-mode",apmmode,0,255);
 				case GET_SET_PARM('c',"32-bit-IO",io32bit,0,3);
 				case     SET_FLAG('C',powermode);
 				case GET_SET_PARM('d',"dma-enable",dma,0,1);
