@@ -18,19 +18,20 @@
 #include <linux/hdreg.h>
 
 extern int verbose;
+extern int prefer_ata12;
 
 /*
  * Taskfile layout for SG_ATA_16 cdb:
  *
  * LBA48:
- * cdb[ 3] = hob_feature
+ * cdb[ 3] = hob_feat
  * cdb[ 5] = hob_nsect
  * cdb[ 7] = hob_lbal
  * cdb[ 9] = hob_lbam
  * cdb[11] = hob_lbah
  *
  * LBA28/LBA48:
- * cdb[ 4] = feature
+ * cdb[ 4] = feat
  * cdb[ 6] = nsect
  * cdb[ 8] = lbal
  * cdb[10] = lbam
@@ -38,34 +39,64 @@ extern int verbose;
  * cdb[13] = device
  * cdb[14] = command
  *
+ * Taskfile layout for SG_ATA_12 cdb:
+ *
+ * cdb[ 3] = feat
+ * cdb[ 4] = nsect
+ * cdb[ 5] = lbal
+ * cdb[ 6] = lbam
+ * cdb[ 7] = lbah
+ * cdb[ 8] = device
+ * cdb[ 9] = command
+ *
  * dxfer_direction choices:
  *	SG_DXFER_TO_DEV, SG_DXFER_FROM_DEV, SG_DXFER_NONE
  */
 
+static inline int needs_lba48 (__u8 ata_op, __u64 lba, unsigned int nsect)
+{
+	const __u64 lba28_limit = (1<<28) - 1;
+
+	switch (ata_op) {
+		case ATA_OP_READ_PIO_EXT:
+		case ATA_OP_READ_DMA_EXT:
+		case ATA_OP_WRITE_PIO_EXT:
+		case ATA_OP_WRITE_DMA_EXT:
+		case ATA_OP_READ_VERIFY_EXT:
+		case ATA_OP_WRITE_UNC_EXT:
+		case ATA_OP_READ_NATIVE_MAX_EXT:
+		case ATA_OP_SET_MAX_EXT:
+		case ATA_OP_FLUSHCACHE_EXT:
+			return 1;
+	}
+	if (lba >= lba28_limit)
+		return 1;
+	if (nsect) {
+		if (nsect > 0xff)
+			return 1;
+		if ((lba + nsect - 1) >= lba28_limit)
+			return 1;
+	}
+	return 0;
+}
+
 void tf_init (struct ata_tf *tf, __u8 ata_op, __u64 lba, unsigned int nsect)
 {
-	const __u64 lba28_mask = 0x0fffffff;
-
 	memset(tf, 0, sizeof(*tf));
-	tf->command = ata_op;
-	tf->dev     = ATA_USING_LBA;
-	if (lba) {
-		tf->lob.lbal = lba;
-		tf->lob.lbam = lba >>  8;
-		tf->lob.lbah = lba >> 16;
-		if ((lba & ~lba28_mask) == 0) {
-			tf->dev |= (lba >> 24) & 0x0f;
-		} else {
-			tf->hob.lbal = lba >> 24;
-			tf->hob.lbam = lba >> 32;
-			tf->hob.lbah = lba >> 40;
-			tf->is_lba48 = 1;
-		}
-	}
-	if (nsect) {
-		tf->lob.nsect = nsect;
-		if ((lba & ~lba28_mask) != 0)
-			tf->hob.nsect = nsect >> 8;
+	tf->command  = ata_op;
+	tf->dev      = ATA_USING_LBA;
+	tf->lob.lbal = lba;
+	tf->lob.lbam = lba >>  8;
+	tf->lob.lbah = lba >> 16;
+	tf->lob.nsect = nsect;
+	if (needs_lba48(ata_op, lba, nsect)) {
+		tf->is_lba48 = 1;
+		tf->hob.nsect = nsect >> 8;
+		tf->hob.lbal = lba >> 24;
+		tf->hob.lbam = lba >> 32;
+		tf->hob.lbah = lba >> 40;
+	} else {
+		tf->dev |= (lba >> 24) & 0x0f;
 	}
 }
 
@@ -110,14 +141,7 @@ static void dump_bytes (const char *prefix, unsigned char *p, int len)
 	fprintf(stderr, "\n");
 }
 
-static int bad_sense (unsigned char *sb, int len)
-{
-	if (verbose)
-		dump_bytes("SG_IO: bad/missing ATA_16 sense data:", sb, len);
-	return EIO;
-}
-
-int sg16 (int fd, int rw, struct ata_tf *tf,
+int sg16 (int fd, int rw, int dma, struct ata_tf *tf,
 	void *data, unsigned int data_bytes, unsigned int timeout_secs)
 {
 	unsigned char cdb[SG_ATA_16_LEN];
@@ -125,38 +149,52 @@ int sg16 (int fd, int rw, struct ata_tf *tf,
 	struct scsi_sg_io_hdr io_hdr;
 
 	memset(&cdb, 0, sizeof(cdb));
-	cdb[ 0] = SG_ATA_16;
-	cdb[ 1] = data ? (rw ? SG_ATA_PROTO_PIO_OUT : SG_ATA_PROTO_PIO_IN) : SG_ATA_PROTO_NON_DATA;
-#if 0
-	if (data_bytes && (data_bytes % 512) == 0) {
-		cdb[1] = SG_ATA_PROTO_DMA;
+	memset(&sb,     0, sizeof(sb));
+	memset(&io_hdr, 0, sizeof(struct scsi_sg_io_hdr));
+
+	if (dma) {
+		//cdb[1] = data ? (rw ? SG_ATA_PROTO_UDMA_OUT : SG_ATA_PROTO_UDMA_IN) : SG_ATA_PROTO_NON_DATA;
+		cdb[1] = data ? SG_ATA_PROTO_DMA : SG_ATA_PROTO_NON_DATA;
+	} else {
+		cdb[1] = data ? (rw ? SG_ATA_PROTO_PIO_OUT : SG_ATA_PROTO_PIO_IN) : SG_ATA_PROTO_NON_DATA;
 	}
-#endif
 	cdb[ 2] = SG_CDB2_CHECK_COND;
 	if (data) {
 		cdb[2] |= SG_CDB2_TLEN_NSECT | SG_CDB2_TLEN_SECTORS;
 		cdb[2] |= rw ? SG_CDB2_TDIR_TO_DEV : SG_CDB2_TDIR_FROM_DEV;
 	}
-	cdb[ 4] = tf->lob.feat;
-	cdb[ 6] = tf->lob.nsect;
-	cdb[ 8] = tf->lob.lbal;
-	cdb[10] = tf->lob.lbam;
-	cdb[12] = tf->lob.lbah;
-	cdb[13] = tf->dev;
-	cdb[14] = tf->command;
-	if (tf->is_lba48) {
-		cdb[ 1] |= SG_ATA_LBA48;
-		cdb[ 3]  = tf->hob.feat;
-		cdb[ 5]  = tf->hob.nsect;
-		cdb[ 7]  = tf->hob.lbal;
-		cdb[ 9]  = tf->hob.lbam;
-		cdb[11]  = tf->hob.lbah;
+
+	if (!prefer_ata12 || tf->is_lba48) {	//FIXME
+		cdb[ 0] = SG_ATA_16;
+		cdb[ 4] = tf->lob.feat;
+		cdb[ 6] = tf->lob.nsect;
+		cdb[ 8] = tf->lob.lbal;
+		cdb[10] = tf->lob.lbam;
+		cdb[12] = tf->lob.lbah;
+		cdb[13] = tf->dev;
+		cdb[14] = tf->command;
+		if (tf->is_lba48) {
+			cdb[ 1] |= SG_ATA_LBA48;
+			cdb[ 3]  = tf->hob.feat;
+			cdb[ 5]  = tf->hob.nsect;
+			cdb[ 7]  = tf->hob.lbal;
+			cdb[ 9]  = tf->hob.lbam;
+			cdb[11]  = tf->hob.lbah;
+		}
+		io_hdr.cmd_len = SG_ATA_16_LEN;
+	} else {
+		cdb[ 0] = SG_ATA_12;
+		cdb[ 3] = tf->lob.feat;
+		cdb[ 4] = tf->lob.nsect;
+		cdb[ 5] = tf->lob.lbal;
+		cdb[ 6] = tf->lob.lbam;
+		cdb[ 7] = tf->lob.lbah;
+		cdb[ 8] = tf->dev;
+		cdb[ 9] = tf->command;
+		io_hdr.cmd_len = SG_ATA_12_LEN;
 	}
 
-	memset(&sb,     0, sizeof(sb));
-	memset(&io_hdr, 0, sizeof(struct scsi_sg_io_hdr));
 	io_hdr.interface_id	= 'S';
-	io_hdr.cmd_len		= SG_ATA_16_LEN;
 	io_hdr.mx_sb_len	= sizeof(sb);
 	io_hdr.dxfer_direction	= data ? (rw ? SG_DXFER_TO_DEV : SG_DXFER_FROM_DEV) : SG_DXFER_NONE;
 	io_hdr.dxfer_len	= data ? data_bytes : 0;
@@ -174,28 +212,35 @@ int sg16 (int fd, int rw, struct ata_tf *tf,
 		return -1;	/* SG_IO not supported */
 	}
 	if (verbose)
-		fprintf(stderr, "SG_IO: ATA_16 status=0x%x, host_status=0x%x, driver_status=0x%x\n",
-			io_hdr.status, io_hdr.host_status, io_hdr.driver_status);
+		fprintf(stderr, "SG_IO: ATA_%u status=0x%x, host_status=0x%x, driver_status=0x%x\n",
+			io_hdr.cmd_len, io_hdr.status, io_hdr.host_status, io_hdr.driver_status);
 
 	if (io_hdr.host_status || io_hdr.driver_status != SG_DRIVER_SENSE
 	 || (io_hdr.status && io_hdr.status != SG_CHECK_CONDITION))
 	{
-	  	errno = EIO;
+		if (verbose)
+			fprintf(stderr, "SG_IO: bad response (not CHECK_CONDITION)\n");
+	  	errno = EBADE;
 		return -1;
 	}
-	if (verbose)
-		fprintf(stderr, "SG_IO: sb[] = {%02x %02x %02x %02x %02x %02x %02x %02x}\n",
-			sb[0], sb[1], sb[2], sb[3], sb[4], sb[5], sb[6], sb[7]); 
-
-	if (sb[0] != 0x72 || sb[7] < 14)
-		return bad_sense(sb, sizeof(sb));
 
 	desc = sb + 8;
-	if (verbose)
-		fprintf(stderr, "SG_IO: desc[] = {%02x %02x .. }\n", desc[0], desc[1]);
+	if (sb[0] != 0x72 || sb[7] < 14 || desc[0] != 0x09 || desc[1] < 0x0c) {
+		if (verbose)
+			dump_bytes("SG_IO: bad/missing sense data, sb[]", sb, sizeof(sb));
+		errno = EBADE;
+		return -1;
+	}
 
-	if (desc[0] != 9 || desc[1] < 12)
-		return bad_sense(sb, sizeof(sb));
+	if (verbose)
+		dump_bytes("SG_IO: sb[]", sb, sizeof(sb));
+
+	if (verbose) {
+		int len = desc[1], maxlen = sizeof(sb) - 8 - 2;
+		if (len > maxlen)
+			len = maxlen;
+		dump_bytes("SG_IO: desc[]", desc, len);
+	}
 
 	tf->is_lba48  = desc[ 2] & 1;
 	tf->error     = desc[ 3];
@@ -205,20 +250,31 @@ int sg16 (int fd, int rw, struct ata_tf *tf,
 	tf->lob.lbah  = desc[11];
 	tf->dev       = desc[12];
 	tf->status    = desc[13];
+	tf->hob.feat  = 0;
 	if (tf->is_lba48) {
 		tf->hob.nsect = desc[ 4];
 		tf->hob.lbal  = desc[ 6];
 		tf->hob.lbam  = desc[ 8];
 		tf->hob.lbah  = desc[10];
+	} else {
+		tf->hob.nsect = 0;
+		tf->hob.lbal  = 0;
+		tf->hob.lbam  = 0;
+		tf->hob.lbah  = 0;
 	}
+
 	if (verbose)
-		fprintf(stderr, "      ATA_16 tf->status=0x%02x tf->error=0x%02x\n", tf->status, tf->error);
-#if 0 //FIXME: we have to check this somewhere !!
+		fprintf(stderr, "      ATA_%u stat=%02x err=%02x nsect=%02x lbal=%02x lbam=%02x lbah=%02x dev=%02x\n",
+				io_hdr.cmd_len, tf->status, tf->error, tf->lob.nsect, tf->lob.lbal, tf->lob.lbam, tf->lob.lbah, tf->dev);
+
 	if (tf->status & (ATA_STAT_ERR | ATA_STAT_DRQ)) {
+		if (verbose) {
+			fprintf(stderr, "I/O error, ata_op=0x%02x ata_status=0x%02x ata_error=0x%02x\n",
+				tf->command, tf->status, tf->error);
+		}
 		errno = EIO;
 		return -1;
 	}
-#endif
 	return 0;
 }
 
@@ -251,23 +307,17 @@ int do_drive_cmd (int fd, unsigned char *args)
 		tf.lob.lbah  = 0xc2;
 	}
 
-	rc = sg16(fd, SG_READ, &tf, data, data_bytes, 0);
-	if (rc) {
-		if (rc == -1)
+	rc = sg16(fd, SG_READ, SG_PIO, &tf, data, data_bytes, 0);
+	if (rc == -1) {
+		if (errno == EINVAL || errno == ENODEV)
 			goto use_legacy_ioctl;
-		errno = rc;
-		rc = -1;
-	} else if (tf.status & (ATA_STAT_ERR | ATA_STAT_DRQ)) {
-		if (verbose)
-			fprintf(stderr, "I/O error, ata_op=0x%02x ata_status=0x%02x ata_error=0x%02x\n",
-				tf.command, tf.status, tf.error);
-		errno = EIO;
-		rc = -1;
 	}
 
-	args[0] = tf.status;
-	args[1] = tf.error;
-	args[2] = tf.lob.nsect;
+	if (rc == 0 || errno == EIO) {
+		args[0] = tf.status;
+		args[1] = tf.error;
+		args[2] = tf.lob.nsect;
+	}
 	return rc;
 
 use_legacy_ioctl:
@@ -317,32 +367,26 @@ int do_taskfile_cmd (int fd, struct hdio_taskfile *r, unsigned int timeout_secs)
 			break;
 	}
 
-	rc = sg16(fd, rw, &tf, data, data_bytes, timeout_secs);
-	if (rc) {
-		if (rc == -1)
+	rc = sg16(fd, rw, SG_PIO, &tf, data, data_bytes, timeout_secs);
+	if (rc == -1) {
+		if (errno == EINVAL || errno == ENODEV)
 			goto use_legacy_ioctl;
-		errno = rc;
-		rc = -1;
-	} else if (tf.status & (ATA_STAT_ERR | ATA_STAT_DRQ)) {
-		if (verbose)
-			fprintf(stderr, "I/O error, ata_op=0x%02x ata_status=0x%02x ata_error=0x%02x\n",
-				tf.command, tf.status, tf.error);
-		errno = EIO;
-		rc = -1;
 	}
 
-	if (r->iflags.b.feat)		r->lob.feat  = tf.error;
-	if (r->iflags.b.lbal)		r->lob.lbal  = tf.lob.lbal;
-	if (r->iflags.b.nsect)		r->lob.nsect = tf.lob.nsect;
-	if (r->iflags.b.lbam)		r->lob.lbam  = tf.lob.lbam;
-	if (r->iflags.b.lbah)		r->lob.lbah  = tf.lob.lbah;
-	if (r->iflags.b.dev)		r->lob.dev   = tf.dev;
-	if (r->iflags.b.command)	r->lob.command = tf.status;
-	if (r->iflags.b.hob_feat)	r->hob.feat  = tf.hob.feat;
-	if (r->iflags.b.hob_lbal)	r->hob.lbal  = tf.hob.lbal;
-	if (r->iflags.b.hob_nsect)	r->hob.nsect = tf.hob.nsect;
-	if (r->iflags.b.hob_lbam)	r->hob.lbam  = tf.hob.lbam;
-	if (r->iflags.b.hob_lbah)	r->hob.lbah  = tf.hob.lbah;
+	if (rc == 0 || errno == EIO) {
+		if (r->iflags.b.feat)		r->lob.feat  = tf.error;
+		if (r->iflags.b.lbal)		r->lob.lbal  = tf.lob.lbal;
+		if (r->iflags.b.nsect)		r->lob.nsect = tf.lob.nsect;
+		if (r->iflags.b.lbam)		r->lob.lbam  = tf.lob.lbam;
+		if (r->iflags.b.lbah)		r->lob.lbah  = tf.lob.lbah;
+		if (r->iflags.b.dev)		r->lob.dev   = tf.dev;
+		if (r->iflags.b.command)	r->lob.command = tf.status;
+		if (r->iflags.b.hob_feat)	r->hob.feat  = tf.hob.feat;
+		if (r->iflags.b.hob_lbal)	r->hob.lbal  = tf.hob.lbal;
+		if (r->iflags.b.hob_nsect)	r->hob.nsect = tf.hob.nsect;
+		if (r->iflags.b.hob_lbam)	r->hob.lbam  = tf.hob.lbam;
+		if (r->iflags.b.hob_lbah)	r->hob.lbah  = tf.hob.lbah;
+	}
 	return rc;
 
 use_legacy_ioctl:
@@ -381,8 +425,6 @@ use_legacy_ioctl:
 void init_hdio_taskfile (struct hdio_taskfile *r, __u8 ata_op, int rw, int force_lba48,
 				__u64 lba, unsigned int nsect, int data_bytes)
 {
-	const __u64 lba28_mask = 0x0fffffff;
-
 	memset(r, 0, sizeof(struct hdio_taskfile) + data_bytes);
 	if (!data_bytes) {
 		r->dphase  = TASKFILE_DPHASE_NONE;
@@ -413,9 +455,7 @@ void init_hdio_taskfile (struct hdio_taskfile *r, __u8 ata_op, int rw, int force
 	r->lob.lbah  = lba >> 16;
 	r->lob.dev   = ATA_USING_LBA;
 
-	if ((lba & ~lba28_mask) == 0 && nsect <= 256 && !force_lba48) {
-		r->lob.dev |= (lba >> 24) & 0x0f;
-	} else {
+	if (needs_lba48(ata_op, lba, nsect) || force_lba48) {
 		r->hob.nsect = nsect >>  8;
 		r->hob.lbal  = lba   >> 24;
 		r->hob.lbam  = lba   >> 32;
@@ -424,5 +464,7 @@ void init_hdio_taskfile (struct hdio_taskfile *r, __u8 ata_op, int rw, int force
 		r->oflags.b.hob_lbal  = 1;
 		r->oflags.b.hob_lbam  = 1;
 		r->oflags.b.hob_lbah  = 1;
+	} else {
+		r->lob.dev |= (lba >> 24) & 0x0f;
 	}
 }
